/**
 * Copyright (c) CPU and Fundamental Software Research Center, CAS
 *
 * This software is provided by the copyright holders - CPU and Fundamental Software Research Center,
 * CAS. The above copyright is protected by copyright laws in all countries and the Universal Copyright
 * Convention. CPU and Fundamental Software Research Center, CAS will reserve the right to
 * take any necessary legal means against the unauthorized use of this software or any use that appears
 * outside of CPU and Fundamental Softwre Research Center's intellectual property declaration.
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
**/

var apkRT = {};
var apkmanagernode = null;

var EventEmitter = require('events').EventEmitter;
var emitter = new EventEmitter();
var apkLauncherInfo = null;
var installRemindText = chrome.i18n.getMessage('installAlertMsg');//"待安装应用非本系统认证的安卓应用，无法确保可正常安装或全部功能可用，是否继续？";

function beginsWith(str, prefix) {
    return str.substring(0, prefix.length) === prefix;
}

// TODO: Move the path setting into DEB installation.
if (typeof nw != "undefined" && typeof nw.require != "undefined" && typeof nw.process != "undefined" &&  nw.process.__nwjs == 1) {
    apkRT.ver = 13;
    apkRT.require = nw.require;
    apkRT.process = nw.process;
    var path = apkRT.require('path');
    apkRT.arcPluginDir = path.join(nw.__dirname, '..', 'apklauncher', 'arc');
    apkRT.arcExecPath = apkRT.process.execPath;
} else if (typeof process != "undefined" && process.__node_webkit == 1 && typeof require == "function") {
    apkRT.ver = 12;
    apkRT.require = require;
    apkRT.process = process;
    var path = apkRT.require('path');
    apkRT.arcPluginDir = path.join(apkRT.process.cwd(), '..', 'apklauncher', 'arc');
    apkRT.arcExecPath = path.join(apkRT.process.cwd(), '..', 'nw13_custom', 'nw');
}

function InitTexts() {
  document.getElementById('main_title').innerText = chrome.i18n.getMessage('title');
  document.getElementById('alert_title_span').innerText = chrome.i18n.getMessage('alert');
  document.getElementById('alert_reminder_label').innerText = chrome.i18n.getMessage('alertMsg');
  document.getElementById('alert_ok_lable').innerText = chrome.i18n.getMessage('ok');
  document.getElementById('alert_cancel_lable').innerText = chrome.i18n.getMessage('cancel');
}

function onInit() {
  console.log("APK Installer start!");
  apkmanagernode = apkRT.require("arc-apk-manager");
  var doc = document.documentElement;
  var output = document.getElementById("maindiv");

  InitTexts() ;

  var buttons = [
    'alert_close',
    'alert_ok',
    'alert_cancel'
  ];
  for (var i = 0; i < buttons.length; ++i) {
    var button = document.getElementById(buttons[i]);
    button.addEventListener('click', function(e) {
      if (typeof apkRT.ver == "undefined") {
        return;
      }
      if (e.currentTarget.id == 'alert_ok') {
        var alert_checkbox = document.getElementById('alert_reminder');
        var checked = false;
        if (alert_checkbox && alert_checkbox.checked) {
          checked = true;
        }
        pNode = e.currentTarget.parentNode;
        idNode = pNode.childNodes[0];
        var ids = idNode.id.split(";");
        var nodeID = ids[0];
        if (nodeID == "ALERT_INSTALL_REMINDER") {
          var type = ids[1];
          if (type == "installerRemind") {
            emitter.emit('installerRemind');
          }
          if (checked) {
            writeReminderInfo(type, false);
          }
          hideAlert();
         }
       } else if (e.currentTarget.id == 'alert_close' || e.currentTarget.id == 'alert_cancel') {
          hideAlert();
          window.close();
       }
     });
    }

  if (typeof apkRT.ver == "undefined"){
    console.log("Error: No nw support. You should contact admin to sovle this problem.");
    output.innerText = "apkRT " + chrome.i18n.getMessage('error');
    return;
  }

  showInstallRemind();
}

function showInstallRemind() {
  apkmanagernode.getApkLauncherInfo(function(launcherInfo) {
    apkLauncherInfo = launcherInfo;
    if (apkLauncherInfo["needInstallRemind"]) {
      showAddAppReminder("installerRemind", installRemindText, false);
      emitter.once('installerRemind', function() {
        startInstall();
      });
    } else {
      startInstall();
    }
  });
}

function startInstall() {
  var output = document.getElementById("maindiv");
  if (apkRT.process.env['APKTOBEINSTALLED'] != undefined) {
      apkfilepath = apkRT.process.env['APKTOBEINSTALLED']
      output.innerText = chrome.i18n.getMessage('installing') + apkfilepath;
      apkmanagernode.init(null, apkRT.arcExecPath);
      InitApkList(function() {
      apkmanagernode.addApkFile(apkfilepath, function(appID, appName){
        output.innerText = chrome.i18n.getMessage('installing') + appName + "...";
      }, function(jsonInfo){
        console.dir(jsonInfo);
          if (jsonInfo.overrideAppID) {
            output.innerText = jsonInfo.arc_metadata.name + chrome.i18n.getMessage('overrideOk');
          } else {
            output.innerText = jsonInfo.arc_metadata.name + chrome.i18n.getMessage('installOk');
          }
      }, function(e){
        console.log(e.stack);
        if (beginsWith(e.message, "EISDIR: illegal operation on a directory")) {
          output.innerText = chrome.i18n.getMessage('dir_op_err');
        } else if(beginsWith(e.message, "APK does not contain")) {
          output.innerText = chrome.i18n.getMessage('apk_err');
        } else {
          output.innerText = e.message;
        }
        });
      });
    } else {
      output.innerText = chrome.i18n.getMessage('closeAlert1', ['3']);
      setTimeout(function(){
        output.innerText = chrome.i18n.getMessage('closeAlert2', ['2']);
      }, 1000);
      setTimeout(function(){
        output.innerText = chrome.i18n.getMessage('closeAlert2',['1']);
      }, 2000);
      setTimeout(function(){
        window.close();
      }, 3000);
    }
}

//parameter text: text to show
//parameter remind: if show remind checkbox
//parameter type: remind type("dragRemind", "openRemind", "installerRemind", "launchRemind")
function showAddAppReminder(type, text, remind) {
  showAlert("ALERT_INSTALL_REMINDER;" + type, text, false, remind, [chrome.i18n.getMessage('continue')]);
}

function writeReminderInfo(type, remind) {
  var key = "";
  if (type == "installerRemind") {
    key = "needInstallRemind";
  }
  apkLauncherInfo[key] = remind;
  apkmanagernode.saveApkLauncherInfo(apkLauncherInfo, errorCommonHandler);
}

function showAlert(id, message, oneButton, checkbox, buttonTexts) {
  showAlertImp(id, true, message, oneButton, checkbox, buttonTexts);
}

function hideAlert() {
  showAlertImp(null, false, null, null, null);
}

function showAlertImp(id, show, message, oneButton, checkbox, buttonTexts) {
  // console.log("showAlertImp, id=%s, show=%s, msg=%s, oneButton=%s, checkbox=%s",
  //       id, show, message, oneButton, checkbox);
  var alert_box = document.getElementById("alert_box");
  var alert_content = document.getElementById("alert_content");
  var alert_checkbox = document.getElementById("alert_checkbox");
  var alert_ok_btn = document.getElementById("alert_ok");
  var alert_cancel_btn = document.getElementById("alert_cancel");
  var alert_foot = document.getElementById("alert_foot");
  var alert_ok_lable = document.getElementById("alert_ok_lable");
  var alert_cancel_lable = document.getElementById("alert_cancel_lable");

  var idDiv = document.createElement("div");
  idDiv.id = id;
  idDiv.style.display = "none"
  alert_foot.insertBefore(idDiv,alert_foot.childNodes[0]);

  if (show) {
    alert_box.style.display = 'flex';
    alert_content.innerText = message;
  } else {  //hide
    alert_box.style.display = 'none';
  }
  if (!checkbox) {
    alert_checkbox.style.display = 'none';
  } else {
    alert_checkbox.style.display = 'inline-block';
  }
  if (buttonTexts) {
    if (buttonTexts.length == 1) {
      alert_ok_lable.innerText = buttonTexts[0];
    } else if (buttonTexts.length > 1) {
      alert_ok_lable.innerText = buttonTexts[0];
      alert_cancel_lable.innerText = buttonTexts[1];
    }
  } else {
    if (oneButton) {
      alert_ok_lable.innerText = chrome.i18n.getMessage('Ok');
    } else {
      alert_ok_lable.innerText = chrome.i18n.getMessage('Ok');
      alert_cancel_lable.innerText = chrome.i18n.getMessage('Cancel');
    }
  }
  if (oneButton) {
    alert_ok_btn.style.display = 'inline-block';
    alert_cancel_btn.style.display = 'none';
    alert_ok_btn.style.marginLeft = '170px';
  } else {
    alert_ok_btn.style.display = 'inline-block';
    alert_cancel_btn.style.display = 'inline-block';
    alert_ok_btn.style.marginLeft = '90px';
  }
}

function InitApkList(callback) {
  apkmanagernode.getApkList(function(appIDList) {
    return Promise.all(appIDList.map(getApkinfo_promise)).then(function(){
      callback();
    });
  }, function(error) {
    callback();
  });
}

function getApkinfo_promise(appID) {
  return new Promise(function(resolve, reject) {
    apkmanagernode.getApkInfo(appID, function(jsonInfo, isPre) {
      resolve();
    }, function(error) {
      resolve();
    });
  });
}
document.addEventListener('DOMContentLoaded', onInit, false);
