/**
 * Copyright (c) CPU and Fundamental Software Research Center, CAS
 *
 * This software is provided by the copyright holders - CPU and Fundamental Software Research Center,
 * CAS. The above copyright is protected by copyright laws in all countries and the Universal Copyright
 * Convention. CPU and Fundamental Software Research Center, CAS will reserve the right to 
 * take any necessary legal means against the unauthorized use of this software or any use that appears
 * outside of CPU and Fundamental Softwre Research Center's intellectual property declaration.
 * 
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
**/

var fs = require('fs');
var path = require('path');
var util = require('util');
var mime = require("mime");
var apkreader = require('adbkit-apkreader-plus');

var arcPluginDir = null; // arc plugin Directory
var appsDir = null; // app install dir
var appsDataDir = null; // app data running dir
var preInstalledAppsDir = null;
var arcExecPath = null; // run arc-apk program

var xdg_desktop_dir = null;
var xdg_menu_file = null;
var xdg_menu_home_dir = null;
var apk_categrories = 'Network;';

var apkInfoList = {};
var preApkInfoList = {};
var INSTALL_SOCK_PATH = '/tmp/apkrt-install.sock';
var apkLauncherInfoJsonPath = null;

function test(filepath) {
  console.log("dir" + __dirname);
  util.log("dir" + __dirname);
  return 1;
}

function testContext(fn) {
  console.log("This is the context of console");
  fn(__dirname);
  setTimeout(fn, 0, __dirname);
  setTimeout(function() {
    console.log("This is test setTimout console in node.");
  }, 0);
  console.log("This is the context of util");
}

function init(sArcPluginDir, sArcExecPath) {
  console.log("init __dirname %s", __dirname);
  if (sArcExecPath){
    arcExecPath = sArcExecPath;
  }
  var sAppDataDirectory = path.join(process.env['HOME'], '.config', 'APK-Launcher-AppData');
  preInstalledAppsDir =  path.join('/', 'opt', 'apkruntime', 'apps');
  apkLauncherInfoJsonPath = path.join(sAppDataDirectory, 'apklauncher.json');


  try {
    fs.mkdirSync(sAppDataDirectory);
  } catch (e) {}
  appsDir = path.join(sAppDataDirectory, 'apps');
  try {
    fs.mkdirSync(appsDir);
  } catch (e) {}
  appsDataDir = path.join(sAppDataDirectory, 'app-data');
  try {
    fs.mkdirSync(appsDataDir);
  } catch (e) {}
  arcPluginDir = path.join(sAppDataDirectory, 'arcLink');
  if (sArcPluginDir) {
    try {
      var linkPath = fs.readlinkSync(arcPluginDir);
      if (linkPath !== sArcPluginDir) {
        fs.unlinkSync(arcPluginDir);
      }
    } catch (e) {
      console.dir(e);
    }
    try {
      fs.symlinkSync(sArcPluginDir, arcPluginDir);
    } catch (e) {}
  }
  xdg_menu_file = path.join('/', 'usr' , 'share', 'desktop-directories', 'apk-apps.directory');;
  if (!process.env['XDG_DATA_HOME']) {
    xdg_desktop_dir = path.join(process.env['HOME'], '.local', 'share', 'applications');
  }else {
    xdg_desktop_dir = path.join(process.env['XDG_DATA_HOME'], '.local', 'share', 'applications');
  }
  xdg_menu_home_dir = path.join('/', 'usr', 'share', 'desktop-directories');
  try {
    fs.mkdirSync(xdg_desktop_dir );
  } catch(e) {}
  try {
    fs.mkdirSync(xdg_menu_home_dir);
  } catch(e) {}
}

function readApkLauncherInfo_Promise(infoJsonFile) {
  return requestJSON_Promise(infoJsonFile).then(function(jsonInfo) {
    return Promise.resolve(jsonInfo);
  }).catch(function(err) {
    var jsonInfo = {
      "needLaunchRemind": "true",
      "needInstallRemind": "true"
    };
    return Promise.resolve(jsonInfo);
  });
}

function getApkLauncherInfo(successHandler) {
  readApkLauncherInfo_Promise(apkLauncherInfoJsonPath).then(function(jsonInfo) {
    successHandler(jsonInfo);
  });
}

function saveApkLauncherInfo(jsonInfo, errorHandler) {
  fs.writeFile(apkLauncherInfoJsonPath, JSON.stringify(jsonInfo, null, 4), function(err) {
    if (err) {
      setTimeout(errorHandler, 0, err);
      return;
    }
  });
}

function requestJSON_Promise(jsonFileName) {
  return new Promise(function(resolve, reject) {
    fs.readFile(jsonFileName, function(err, data) {
      if (err) {
        reject(new Error(err));
        return;
      }
      try {
        var result = JSON.parse(data);
        resolve(result);
      } catch (e) {
        reject(new Error(e));
      }
    });
  });
}

function copyFile_Promise(source, target) {
  return existsFile_Promise(source).then(function() {
    return new Promise(function(resolve, reject) {
      var cbCalled = false;
      var rd = fs.createReadStream(source, {
        autoClose: true
      });
      rd.on("error", function(err) {
        done(err);
      });
      var wr = fs.createWriteStream(target);
      wr.on("error", function(err) {
        done(err);
      });
      wr.on("close", function(ex) {
        done();
        rd.close();
      });
      rd.pipe(wr);

      function done(err) {
        if (!cbCalled) {
          if (err) {
            reject(err);
            cbCalled = true;
            return;
          }
          resolve();
          cbCalled = true;
        }
      }
    });
  });
}

function deleteFile_Promise(target) {
    return new Promise(function(resolve, reject) {
      fs.unlink(target, function(err, stat) {
        if (err && err.code != 'ENOENT') {
          reject(new Error(err));
        } else {
          resolve();
        }
      });
    });
}

function existsFile_Promise(target) {
    return new Promise(function(resolve, reject) {
      fs.stat(target, function(err, stat) {
        if (err) {
          reject(new Error(err));
        } else {
          resolve(stat);
        }
      });
    });
}

//Deprecated
//Without this statement, menu also is updated.
function xdgMenuUpdate_Promise() {
  return Promise.resolve();
  // return new Promise(function(resolve, reject) {
  //   var spawn = require('child_process').spawn;
  //   var appProcess = spawn('xdg-desktop-menu', ["forceupdate"]);
  //   appProcess.stderr.on('data', function(data) {
  //     console.log('stderr: ' + data);
  //   });
  //   appProcess.on('close', function(code) {
  //     if (code == 0) {
  //       resolve();
  //     } else {
  //       reject(new Error("xdg-desktop-menu update error with code: " + code));
  //     }
  //   });
  // });
}

function xdgMenuUninstall_Promise(desktop_file) {
  return new Promise(function(resolve, reject) {
    var spawn = require('child_process').spawn;
    var appProcess = spawn('xdg-desktop-menu', ["uninstall", "--noupdate", desktop_file]);
    appProcess.stderr.on('data', function(data) {
      console.log('stderr: ' + data);
    });
    appProcess.on('close', function(code) {
      console.log('child process exited with code ' + code);
      if (code == 0) {
        resolve(desktop_file);
      } else {
        reject(new Error("xdg-desktop-menu uninstall error: " + code));
      }
    });
  });
}

function xdgMenuInstall_Promise(desktop_file) {
  return new Promise(function(resolve, reject) {
    var spawn = require('child_process').spawn;
    menu_file = path.join(xdg_menu_home_dir, 'apk-apps.directory');
    // var appProcess = spawn('xdg-desktop-menu', ["install", "--noupdate", menu_file, desktop_file]);
    var appProcess = spawn('xdg-desktop-menu', ["install", "--noupdate", desktop_file]);
    appProcess.stderr.on('data', function(data) {
      console.log('stderr: ' + data);
    });
    appProcess.on('close', function(code) {
      console.log('child process exited with code ' + code);
      if (code == 0) {
        resolve(desktop_file);
      } else {
        reject(new Error("xdg-desktop-menu install error: " + code));
      }
    });
  });
}

function xdgSendToDesktop_Promise(desktop_file) {
  return new Promise(function(resolve, reject) {
    var spawn = require('child_process').spawn;
    var appProcess = spawn('xdg-desktop-icon', ["install", desktop_file]);
    appProcess.stderr.on('data', function(data) {
      console.log('stderr: ' + data);
    });
    appProcess.on('close', function(code) {
      console.log('child process exited with code ' + code);
      if (code == 0) {
        resolve(desktop_file);
      } else {
        reject(new Error("xdg-desktop-icon install error: " + code));
      }
    });
  });
}

function xdgUninstallIconOfDesktop_Promise(desktop_file) {
  return new Promise(function(resolve, reject) {
    var spawn = require('child_process').spawn;
    var appProcess = spawn('xdg-desktop-icon', ["uninstall", desktop_file]);
    appProcess.stderr.on('data', function(data) {
      console.log('stderr: ' + data);
    });
    appProcess.on('close', function(code) {
      console.log('child process exited with code ' + code);
      if (code == 0) {
        resolve(desktop_file);
      } else {
        reject(new Error("xdg-desktop-icon install error: " + code));
      }
    });
  });
}

function endsWith(str, suffix) {
    return str.indexOf(suffix, str.length - suffix.length) !== -1;
}

function beginsWith(str, prefix) {
    return str.substring(0, prefix.length) === prefix;
}

function base64Image(src) {
    var data = fs.readFileSync(src).toString("base64");
    return util.format("url(data:%s;base64,%s)", mime.lookup(src), data);
}

function getApkList(successHandler, errorHandler) {
  if (!appsDir && !preInstalledAppsDir ) {
    errorHandler(new Error("Plugin CRX Dir is null, apkmanager need to be inited"));
    return;
  }
  var apksList = new Array();
  scanAppDir_Promise(preInstalledAppsDir,  true).then( function(preApkList) {
    apksList = apksList.concat(preApkList);
    scanAppDir_Promise(appsDir,  false).then(function(apkList) {
      apksList = apksList.concat(apkList);
      successHandler(apksList);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
      return;
    });
  }).catch(function(err) {
     scanAppDir_Promise(appsDir,  false).then(function(apkList) {
      apksList = apksList.concat(apkList);
      successHandler(apksList);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
      return;
    });
  });
}

function scanAppDir_Promise(dir,  preInstalled) {
  return new Promise(function(resolve, reject) {
    fs.readdir(dir, function(err, files) {
      if (err) {
        reject(err);
        return;
      } else {
        var apkList = new Array();
        for (var i = files.length - 1; i >= 0; i--) {
          if (endsWith(files[i], '.json')) {
          var str = files[i].substring(0, files[i].length - 5);
          apkList.push(str);
          if (preInstalled) {
            preApkInfoList[str] = true;
          }
        }
      };
      resolve(apkList);
    }
    });
  });
}

function runApk_Promise(appID) {
  if (!arcExecPath) {
      return Promise.reject(new Error("No ARC Loader dir path settings"));
      return;
  }
  if (!arcPluginDir) {
      return Promise.reject(new Error("No ARC Plugin dir path setting"));
  }
  var appdir = null;
  if(preApkInfoList[appID] !== undefined) {
     appdir = path.join(preInstalledAppsDir, appID);
  } else {
    appdir = path.join(appsDir, appID);
  }

  return existsFile_Promise(appdir).then(function() {
    return new Promise(function(resolve, reject) {
      var appRunDir = path.join(appsDataDir, appID);
      if (preApkInfoList[appID] == undefined) {
        try {
          var linkPath = fs.readlinkSync(path.join(appdir, "arc"));
          console.log(linkPath);
          if (linkPath !== arcPluginDir) {
            fs.unlinkSync(path.join(appdir, "arc"));
            fs.symlinkSync(arcPluginDir, path.join(appdir, 'arc'));
          }
        } catch (e) {
          console.dir(e);
          fs.unlinkSync(path.join(appdir, "arc"));
          fs.symlinkSync(arcPluginDir, path.join(appdir, 'arc'));
        }
      }
      var spawn = require('child_process').spawn;
      var appProcess = spawn(arcExecPath, ["--user-data-dir=" + appRunDir, appdir]);
      appProcess.stdout.on('data', function(data) {
        console.log('stdout: ' + data);
      });
      appProcess.stderr.on('data', function(data) {
        console.log('stderr: ' + data);
      });
      appProcess.on('close', function(code) {
        console.log('child process exited with code ' + code);
      });
      resolve();
    });
  });
}

function runApk(appID, successHandler, errorHandler) {
  console.log("runApk %s", appID);
  runApk_Promise(appID).then(function() {
      setTimeout(successHandler, 0);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
    });
}

function createDesktopShortcut(appID, successHandler, errorHandler) {
  return createShortcut_Promise(appID)
    .then(xdgSendToDesktop_Promise)
    .then(function() {
      setTimeout(successHandler, 0);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
    });
}

function createShortcut_Promise(appID) {
  console.log("createShortcut_Promise %s", appID);
  if (!arcExecPath){
    return Promise.reject(new Error("No ARC Loader Runtime settings"));
  }
  return Promise.resolve(appID).then(function(appID){
    return new Promise(function(resolve, reject) {
      getApkInfo(appID, function(jsonInfo) {
        resolve(jsonInfo);
      }, function(err) {
        reject(new Error(err));
      });
    });
  }).then(function(jsonInfo){
    return deleteFile_Promise(path.join(xdg_desktop_dir, "apk-" + jsonInfo.appID + '.desktop')).then(function() {
      return jsonInfo;
    });
  }).then(function(jsonInfo){
    return new Promise(function(resolve, reject) {
      var filecontent = '[Desktop Entry]\n' +
        'Version=1.0\n' +
        'Encoding=UTF-8\n' +
        'Terminal=false\n' +
        'Type=Application\n' +
        'Name=' + jsonInfo.arc_metadata.name + '\n' +
        'Categories=Network'  + '\n' +
        'Exec="' + arcExecPath + '" --user-data-dir="' + path.join(appsDataDir, appID) + '" "' + path.join(appsDir, appID) + '"\n' +
        'Icon=' + path.join(appsDir, jsonInfo.appID, 'icon.png') + '\n';
      var desktop_file = path.join(xdg_desktop_dir, "apk-" + jsonInfo.appID + '.desktop');
      fs.writeFile(desktop_file, filecontent, function(err) {
        if (err) {
          reject(new Error(err));
          return;
        }
        resolve(jsonInfo);
      });
    });
  }).then(function(jsonInfo){
    desktop_file = path.join(xdg_desktop_dir, "apk-" + jsonInfo.appID + '.desktop');
    return xdgMenuInstall_Promise(desktop_file);
  });
}

function saveApkInfo(apkJsonInfo, successHandler, errorHandler) {
  var save_files = [
    writeAppInfoToFile_Promise(apkJsonInfo),
    writeAppManifestFile_Promise(apkJsonInfo)
  ];
  Promise.all(save_files).then(function() {
    apkInfoList[apkJsonInfo.appID] = apkJsonInfo;
    setTimeout(successHandler, 0);
  }).catch(function(err) {
    setTimeout(errorHandler, 0, err);
  });
}

function getApkInfo(appID, successHandler, errorHandler) {
  console.log("getApkInfo %s", appID);
  if (apkInfoList[appID] !== undefined){
    successHandler(apkInfoList[appID], preApkInfoList[appID]);
    return;
  }else {
    requestJSON_Promise(path.join(preInstalledAppsDir, appID + '.json')).then(function(jsonInfo){
      preApkInfoList[appID] = true;
      return jsonInfo;
    }).catch(function(err){
      return requestJSON_Promise(path.join(appsDir, appID + '.json'));
    }).then(function(jsonInfo) {
      var realAppDir;
      if (preApkInfoList[appID] !== undefined) {
        realAppDir = preInstalledAppsDir;
      } else {
        realAppDir = appsDir;
      }
      var iconfile = path.join(realAppDir, jsonInfo.appID, 'icon.png');
      console.log("iconfile:" + iconfile);
      fs.exists(iconfile, function(exists) {
        if (exists) {
          var dataUri = base64Image(iconfile);
          jsonInfo['icon'] = dataUri;
        }
        if (jsonInfo['arc_metadata'] == undefined) {
          jsonInfo.arc_metadata = {};
          jsonInfo.arc_metadata.apkList = ["app.apk"];
        }
        if (jsonInfo['enableExternalDirectory'] != undefined) {
          jsonInfo.arc_metadata.enableExternalDirectory = jsonInfo['enableExternalDirectory'];
          delete jsonInfo['enableExternalDirectory'];
        }
        if (jsonInfo['name'] != undefined) {
          jsonInfo.arc_metadata.name = jsonInfo['name'];
          delete jsonInfo['name'];
        }
        if (jsonInfo['packageName'] != undefined) {
          jsonInfo.arc_metadata.packageName = jsonInfo['packageName'];
          delete jsonInfo['packageName'];
        }
        apkInfoList[jsonInfo.appID] = jsonInfo;
        successHandler(jsonInfo, preApkInfoList[appID]);
      });
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
    });
  }
}

function removeDir_Promise(dirPath, doJustClean) {
  return Promise.resolve(dirPath).then(function(dirPath) {
    return new Promise(function(resolve, reject) {
      fs.lstat(dirPath, function(err, stat) {
        if (err) {
          if (err.code == 'ENOENT') {
            resolve(false);
            return;
          }
          reject(new Error(err));
          return;
        }
        if (stat.isSymbolicLink() || stat.isFile()) {
          fs.unlink(dirPath, function(err) {
            if (err) {
              if (err.code == 'ENOENT') {
                resolve(false);
                return;
              }
              reject(new Error(err));
              return;
            }
            resolve(false);
          });
          return;
        } else if (stat.isDirectory()) {
          fs.readdir(dirPath, function(err, files) {
            if (err) {
              reject(new Error(err));
              return;
            }
            for (var i = files.length - 1; i >= 0; i--) {
              files[i] = path.join(dirPath, files[i]);
            };
            Promise.all(files.map(removeDir_Promise)).then(function() {
              resolve(true);
            }).catch(function(error) {
              reject(error);
            });
          });
        }
      });
    })
  }).then(function(exists) {
    if (exists && doJustClean !== true) {
      return new Promise(function(resolve, reject) {
        fs.rmdir(dirPath, function(err) {
          if (err) {
            reject(new Error(err));
            return;
          }
          resolve();
        });
      });
    }
  });
}

function deleteAll_promise() {
  var idArr = Object.keys(apkInfoList).map(function(key) {
    if (preApkInfoList[apkInfoList[key].appID] !== undefined) {
      return "";
    }
    return apkInfoList[key].appID;
  });
  return Promise.all(idArr.map(deleteApk_Promise)).then(function() {
    return removeDir_Promise(appsDir, true).then(function(){
      return removeDir_Promise(appsDataDir, true).catch(function(err) {
        console.log('ignore datapath delete error:' + err);
        console.dir(err);
      });
    });
  }).then(function() {
    return new Promise(function(resolve, reject) {
      fs.readdir(xdg_desktop_dir, function(err, files) {
        if (err) {
          reject(new Error(err));
          return;
        }
        var fileArr = files.filter(function(file) {
          return beginsWith(file, 'apk-') && endsWith(file, '.desktop');
        }).map(function(file) {
          return path.join(xdg_desktop_dir, file);
        });
        Promise.all(fileArr.map(deleteFile_Promise)).then(function() {
          resolve();
        }).catch(function(error) {
          reject(error);
        });
      });
    }).then(function(){
      //TODO: Delete desktop-directory file in HOME dir.
    });
  })
}

function deleteAll(successHandler, errorHandler) {
  console.log("deleteAll");
  deleteAll_promise().then(function() {
      setTimeout(successHandler, 0);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
    })
}

function deleteApkData_Promise(appID) {
  console.log("deleteApkData_Promise %s", appID);
  return removeDir_Promise(path.join(appsDataDir, appID));
}

function deleteApk_Promise(appID, isDeleteData) {
  console.log("deleteApk_Promise %s", appID);
  if (appID == undefined || appID == ""){
    return Promise.resolve();
  }
  if (preApkInfoList[appID] !== undefined) {
    return Promise.reject(new Error("PreApp can not be deleted."));
  }
  if (apkInfoList[appID] !== undefined){
   delete apkInfoList[appID];
  }
  var appdir = path.join(appsDir, appID);
  return removeDir_Promise(appdir).then(function() {
    return deleteFile_Promise(appdir + '.json');
  }).then(function() {
    desktop_file = path.join(xdg_desktop_dir, "apk-" + appID + '.desktop');
    return xdgUninstallIconOfDesktop_Promise(desktop_file)
      .then(xdgMenuUninstall_Promise(desktop_file));
  }).then(function() {
    return deleteFile_Promise(path.join(xdg_desktop_dir, "apk-" + appID + '.desktop')).catch(function(err) {
      console.log('ignore datapath delete error:' + err);
      console.dir(err);
    });
  }).then(function() {
    return deleteApkData_Promise(appID).then(function() {
      console.log("deleteApkData_Promise sucessfully appID: %s", appID);
    }).catch(function(err) {
      console.log('ignore datapath delete error:' + err);
      console.dir(err);
    });
  });
}

function deleteApk(appID, isDeleteData, successHandler, errorHandler) {
  console.log("deleteApk %s ", appID);
  deleteApk_Promise(appID, isDeleteData)
    .then(function() {
      setTimeout(successHandler, 0);
    }).catch(function(err) {
      setTimeout(errorHandler, 0, err);
    });
}

function writeAppInfoToFile_Promise(apkJsonInfo) {
  return new Promise(function(resolve, reject) {
    var icondata = apkJsonInfo['icon'];
    if (icondata != undefined) {
      delete apkJsonInfo['icon'];
    }
    fs.writeFile(apkJsonInfo.appDir + '.json', JSON.stringify(apkJsonInfo, null, 4), function(err) {
      if (icondata != undefined) {
        apkJsonInfo['icon'] = icondata;
        icondata = null;
      }
      if (err) {
        reject(new Error(err));
        return;
      }
      resolve(apkJsonInfo);
    });
  })
}

function writeAppManifestFile_Promise(apkJsonInfo) {
  return new Promise(function(resolve, reject) {
    var manifestJSON = {
      name: apkJsonInfo.arc_metadata.name,
      manifest_version: 2,
      "chromium-args": "--no-sandbox --enable-webgl --ignore-gpu-blacklist --always-enable-dev-tools  --enable-devtools-experiments",
      version: apkJsonInfo.version,
      app: {
        background: {
          page: "main.html"
        }
      },
      "single-instance": true,
      icons: {
        "16": "icon.png",
        "128": "icon.png"
      },
      permissions: [
        "node",
        "gcm", {
          "fileSystem": ["write", "retainEntries", "directory"]
        }, {
          "socket": ["tcp-connect", "tcp-listen", "udp-bind", "udp-send-to", "resolve-host"]
        },
        "unlimitedStorage",
        "notifications",
        "clipboardRead",
        "clipboardWrite",
        "identity",
        "accessibilityFeatures.read",
        "https://clients2.google.com/",
        "videoCapture",
        "audioCapture",
        "identity.email",
        "fullscreen",
        "overrideEscFullscreen",
        "alarms",
        "storage",
        "geolocation"
      ],
      "requirements": {
        "3D": {
          "features": ["webgl"]
        }
      },
      "key": "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCwTjmSbBD9uFJEUR7jMYAL1sPXqCNe7ljIb6dheKIYjkNtLt1YuiAw4Pul/PO+aYtm8tsbFnsNZ73AxhMX9aPxuzwlAak/iN6atxv7h8mwOg6HiHdxTDnOPrCv7fEud5aUFaq7sqDhiqwMpJoIOa1w7c299z3oXIrG1FGEkid1IQIDAQAB",
      arc_metadata: apkJsonInfo.arc_metadata
    };

    fs.writeFile(path.join(apkJsonInfo.appDir, 'manifest.json'), JSON.stringify(manifestJSON, null, 4), function(err) {
      if (err) {
        reject(new Error(err));
        return;
      }
      resolve(apkJsonInfo);
    });
  })
}

function installApkDir_promise(apkJsonInfo) {
  return writeAppInfoToFile_Promise(apkJsonInfo).then(function(apkJsonInfo){
    return new Promise(function(resolve, reject) {
      fs.mkdir(apkJsonInfo.appDir, function(err) {
        if (err) {
          if (err.code === 'EEXIST') {
            removeDir_Promise(apkJsonInfo.appDir, true).then(function(){
              resolve(apkJsonInfo);
            }).catch(function(error){
              reject(new Error(error));
            });
          }else {
            reject(new Error(err));
          }
        } else {
          resolve(apkJsonInfo);
        }
      });
    });
  }).then(function(apkJsonInfo){
        //Install Dir
        fs.symlinkSync(arcPluginDir, path.join(apkJsonInfo.appDir, 'arc'));
        if (apkJsonInfo['iconpath'] !== undefined) {
          return new Promise(function(resolve, reject) {
            var reader = apkreader.readFile(apkJsonInfo.rawFilePath);
            var iconpath = apkJsonInfo['iconpath'];
            delete apkJsonInfo["iconpath"];
            reader.extractFileTo(iconpath, apkJsonInfo.appDir);
            delete reader;
            fs.renameSync(path.join(apkJsonInfo.appDir, path.basename(iconpath)), path.join(apkJsonInfo.appDir, 'icon.png'));
            apkJsonInfo['icon'] = base64Image(path.join(apkJsonInfo.appDir, 'icon.png'));
            resolve(apkJsonInfo);
          });
        }else {
          return copyFile_Promise(path.join(__dirname, 'icon.png'), path.join(apkJsonInfo.appDir, 'icon.png')).then(function(){
            apkJsonInfo['icon'] = base64Image(path.join(apkJsonInfo.appDir, 'icon.png'));
            return apkJsonInfo;
          });
        }
  }).then(writeAppManifestFile_Promise).then(function(apkJsonInfo){
      return new Promise(function(resolve, reject){
        var mainhtml = '<!DOCTYPE html> <meta charset="utf-8"> <script src="arc/metadata.js"></script> <script src="arc/common.js"></script> <script src="arc/clipboard.js"></script> <script src="arc/child_plugin.js"></script> <script src="arc/crash_reporter.js"></script> <script src="arc/gcm.js"></script> <script src="arc/promise_wrap.js"></script> <script src="arc/filesystem_background.js"></script> <script src="arc/dex2oat.js"></script> <script src="arc/uma.js"></script> <script src="arc/main.js"></script><div id="childplugindiv"></div>';
        fs.writeFileSync(path.join(apkJsonInfo.appDir, 'main.html'), mainhtml);

        fs.mkdirSync(path.join(apkJsonInfo.appDir, 'vendor'));
        fs.mkdirSync(path.join(apkJsonInfo.appDir, 'vendor', 'chromium'));
        fs.mkdirSync(path.join(apkJsonInfo.appDir, 'vendor', 'chromium', 'crx'));

        resolve(apkJsonInfo);
      });
  }).then(function(apkJsonInfo){
    return copyFile_Promise(apkJsonInfo.rawFilePath, path.join(apkJsonInfo.appDir, 'vendor', 'chromium', 'crx', 'app.apk')).then(function(){
      return apkJsonInfo;
    });
  }).catch(function(error){
    return Promise.reject(new Error(error));
  });
}

function Notify_apklauncher_Promise(message) {
  return new Promise(function(resolve, reject) {
    var net = require('net');
    const fs = require('fs');
    if (fs.existsSync(INSTALL_SOCK_PATH)) {
      var client = net.connect({path: INSTALL_SOCK_PATH}, function() {
        console.log('connect to server.');
        client.write(message);
        client.end();
        resolve();
      });

      client.on('error', function(err) {
        reject(new Error(err));
      });
    } else {
      reject(new Error('no server.'))
    }
  });
}

function addApkFile(filepath, startHandler, successHandler, errorHandler) {
  console.log("addApkFile %s", filepath);
  Promise.resolve(filepath).then(function(filepath){
    var apkFilepath = filepath;
    var appFileName = path.basename(apkFilepath);

    var reader = apkreader.readFile(apkFilepath);
    var manifest = reader.readManifestSync();
    try {
      var resource = reader.readResourcesSync();
    } catch(error) {
      setTimeout(errorHandler, 0, error);
      if (beginsWith(error.message, "Unexpected chunk Type")) {
        return Promise.reject(new Error(chrome.i18n.getMessage('ErrParseResource')));
      }
      return Promise.reject(new Error("CancelAppInstall"));
    }

    // {
    //     "apkList": [
    //       "LinpackForAndroid_v1.2.8.apk"
    //     ],
    //     "enableExternalDirectory": false,
    //     "name": "Linpack for Android",
    //     "packageName": "com.greenecomputing.linpack",
    //     "version": "1.2.8"
    //   }

    var apkjson = {};
    apkjson.arc_metadata = {
      apkList: ["app.apk"],
      enableExternalDirectory: false,
      name: appFileName,
      packageName: manifest.package,
      appID: manifest.package + '_' + manifest.versionCode
    };
    apkjson.rawFilePath = apkFilepath;
    apkjson.version = manifest.versionName;

    apkjson.appID = manifest.package + '_' + manifest.versionCode;
    apkjson.appDir = path.join(appsDir, apkjson.appID);
    var preExist = fs.existsSync(path.join(preInstalledAppsDir, apkjson.appID));
    var exists = fs.existsSync(apkjson.appDir)  || preExist;
    if (exists){
      if (preExist) {
        throw new Error(chrome.i18n.getMessage('existAlert'));
      } else {
        throw new Error(chrome.i18n.getMessage('existAlert'));
      }
      return;
    }

    if (manifest.application.label != null && manifest.application.label.substring(0, 13) == "resourceId:0x") {
      var tmpname = resource.getResourceById(parseInt(manifest.application.label.substring(11)), 'zh_CN');
      if (tmpname.substring(0, 1) != '@') {
        apkjson.arc_metadata.name = tmpname;
      }
    } else if (manifest.application.label != null) {
      apkjson.arc_metadata.name = manifest.application.label;
    }

    if (manifest.versionName != null && manifest.versionName.substring(0, 13) == "resourceId:0x") {
      var tmpname = resource.getResourceById(parseInt(manifest.versionName.substring(11)), 'zh_CN');
      if (tmpname.substring(0, 1) != '@') {
        apkjson.version = tmpname;
      }
    } else if (manifest.versionName != null) {
      apkjson.version = manifest.versionName;
    }

    if (manifest.application.icon != null && manifest.application.icon.substring(0, 13) == "resourceId:0x") {
      var tmppath = resource.getResourceById(parseInt(manifest.application.icon.substring(11)), 'zh_CN');
      if (tmppath.substring(0, 1) != '@') {
        apkjson.iconpath = tmppath;
      }
    }
    return apkjson;
  }).catch(function(error){
    setTimeout(errorHandler, 0, error);
    // App has been installed Or Apk can't be parsed.
    // Just cancel the later installing step.
    return Promise.reject(new Error("CancelAppInstall"));
  }).then(function(apkJsonInfo){
    setTimeout(startHandler, 0, apkJsonInfo.appID, apkJsonInfo.arc_metadata.name, false);
    return installApkDir_promise(apkJsonInfo).then(function(apkJsonInfo){
      apkJsonInfo.overrideAppID = null;
      return new Promise(function(resolve, reject) {
        for(var key in apkInfoList) {
          if (apkInfoList[key].arc_metadata.packageName == apkJsonInfo.arc_metadata.packageName) {
            apkJsonInfo.overrideAppID = apkInfoList[key].appID;
            deleteApk_Promise(key, true).then(function() {
              resolve(apkJsonInfo);
            }).catch(function(error) {});

            break;
          }
        } //for
        if (apkJsonInfo.overrideAppID == null) {
          resolve(apkJsonInfo);
        }
      }).then(function(apkJsonInfo){
        return createShortcut_Promise(apkJsonInfo.appID).catch(function(err){
          console.log("ignore install shortcut error:" + err);
          console.dir(err);
        }).then(function() {
            var message = "install-ok:" + apkJsonInfo.appID + ";";
            if (apkJsonInfo.overrideAppID) {
              message += "overrideAppID:" + apkJsonInfo.overrideAppID;
            }
            return Notify_apklauncher_Promise(message).catch(function(err) {
            console.log("ignore  error:" + err);
            console.dir(err);
          });
        }).then(function(){
          return apkJsonInfo;
        });
      });
    });
  }).then(function(apkJsonInfo){
    console.log("Finish installed." + apkJsonInfo.appID);
    apkInfoList[apkJsonInfo.appID] = apkJsonInfo;
    successHandler(apkJsonInfo);
  }).catch(function(error){
    if (error.message != "CancelAppInstall") {
      return removeDir_Promise(apkJsonInfo.appDir).catch(function(error) {}).then(function() {
        setTimeout(errorHandler, 0, error);
      });
    }
  });
}

exports.init = init;
exports.addApkFile = addApkFile;
exports.deleteApk = deleteApk;
exports.saveApkInfo = saveApkInfo;
exports.getApkInfo = getApkInfo;
exports.getApkList = getApkList;
exports.getApkLauncherInfo = getApkLauncherInfo;
exports.saveApkLauncherInfo = saveApkLauncherInfo;
exports.runApk = runApk;
exports.createDesktopShortcut = createDesktopShortcut;
exports.deleteAll = deleteAll;

exports.test = test;
exports.testContext = testContext;
