/**
 * Copyright (c) CPU and Fundamental Software Research Center, CAS
 *
 * This software is provided by the copyright holders - CPU and Fundamental Software Research Center,
 * CAS. The above copyright is protected by copyright laws in all countries and the Universal Copyright
 * Convention. CPU and Fundamental Software Research Center, CAS will reserve the right to 
 * take any necessary legal means against the unauthorized use of this software or any use that appears
 * outside of CPU and Fundamental Softwre Research Center's intellectual property declaration.
 * 
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
**/

(function() {
  var ApkReader, BinaryXmlParser, ManifestParser, Zip;

  Zip = require('adm-zip');

  ManifestParser = require('./apkreader/parser/manifest');

  BinaryXmlParser = require('./apkreader/parser/binaryxml');

  ResourcesParser = require('./apkreader/parser/resources');

  ApkReader = (function() {
    var MANIFEST;

    MANIFEST = 'AndroidManifest.xml';

    ApkReader.readFile = function(apk) {
      return new ApkReader(apk);
    };

    function ApkReader(apk) {
      var err;
      this.apk = apk;
      try {
        this.zip = new Zip(this.apk);
      } catch (_error) {
        err = _error;
        if (typeof err === 'string') {
          throw new Error(err);
        } else {
          throw err;
        }
      }
    }

    ApkReader.prototype.readManifestSync = function() {
      var manifest;
      if (manifest = this.zip.getEntry(MANIFEST)) {
        return new ManifestParser(manifest.getData()).parse();
      } else {
        throw new Error("APK does not contain '" + MANIFEST + "'");
      }
    };

    ApkReader.prototype.readXmlSync = function(path) {
      var file;
      if (file = this.zip.getEntry(path)) {
        return new BinaryXmlParser(file.getData()).parse();
      } else {
        throw new Error("APK does not contain '" + path + "'");
      }
    };

    ApkReader.prototype.readResourcesSync = function() {
      var file;
      if (file = this.zip.getEntry('resources.arsc')) {
        return new ResourcesParser(file.getData()).parse();
      } else {
        throw new Error("APK does not contain resources.arsc");
      }
    };

    ApkReader.prototype.extractFileTo = function(entryPath, targetPath) {
      var file;
      if (file = this.zip.getEntry(entryPath)) {
          this.zip.extractEntryTo(entryPath, targetPath, false, true);
      } else {
        throw new Error("APK does not contain '" + path + "'");
      }
    };

    return ApkReader;

  })();

  module.exports = ApkReader;

}).call(this);
